import numpy as np
import torch
import sklearn.metrics as sk_metrics

class Metric():
    '''
    Metrics holds values during accumulation and reduce to stats in the \
        end, should be aware of memory issue when saving all scores with \
        with large datasets. Also hold the best metrics/epoch for the run.
    '''
    
    def __init__(self):
        self.memory = []
        self.best_result = None
        self.best_epoch = None

    def reset_memory(self):
        self.memory = []

    def step_wrap(self, output_data):
        stats = self.step(output_data)
        self.memory.append(stats)

    def reduce_wrap(self, config):
        result = self.reduce(self.memory, config)
        self.update_result(result, config)
        self.reset_memory()
        return result

    def update_result(self, new, config):
        current_epoch = config['data']['current_epoch']
        old = self.best_result
        if self.best_result is None or self.select_better(old, new) == new:
            self.best_result = new
            self.best_epoch = current_epoch

    def get_best(self):
        '''
        Get the best result from current run

        Returns:
            float: best result of this metric
            int: epoch where the best result is reached
        '''
        return self.best_result, self.best_epoch

    def step(self, output_data):
        '''
        Store scores to avoid large RAM usage

        Args:
            output_data (tuple): Output from train function has form: \
                ``loss.item(), inputs, outputs, labels`` or form defined \
                by customized train function.

        Returns:
            list: an array of transformed stats for computation.
        '''
        raise NotImplementedError('Accumulation not implemented')

    def reduce(self, memory, config):
        '''
        Reduce arrays using the format in the accumulation phase

        Args:
            config (dict): Determine global progress, Metric is memoryless.
        '''
        raise NotImplementedError('Reduce not implemented')

    def select_better(self, old, new):
        '''
        Select the better metrics from two metrics

        Args:
            old (float): old metrics data
            new (float): new metrics data

        Returns:
            float: old or new metrics
        '''
        raise NotImplementedError('Reduce not implemented')


class Loss(Metric):

    def step(self, outputs):
        loss, inputs, outputs, labels = outputs
        return loss

    def reduce(self, memory, config):
        return sum(memory) / len(memory)

    def select_better(self, old, new):
        return min(old, new)

class Accuracy(Metric):
    
    def step(self, output_data):
        loss, inputs, outputs, labels = output_data
        return outputs, labels

    def reduce(self, memory, config):
        outputs = [entry[0] for entry in memory]
        labels = [entry[1] for entry in memory]
        outputs = torch.argmax(torch.cat(outputs), dim = 1).cpu().numpy()
        labels = torch.cat(labels).cpu().numpy()
        score = sk_metrics.accuracy_score(labels, outputs)
        return score

    def select_better(self, old, new):
        return max(old, new)

class AUC(Metric):

    def step(self, output_data):
        loss, inputs, outputs, labels = output_data
        return outputs, labels

    def reduce(self, memory, config):
        outputs = [entry[0] for entry in memory]
        labels = [entry[1] for entry in memory]
        outputs = self.softmax(torch.cat(outputs).cpu().numpy())[:, 1]
        labels = torch.cat(labels).cpu().numpy()
        score = sk_metrics.roc_auc_score(labels, outputs)
        return score

    def softmax(self, x):
        return np.exp(x) / np.exp(x).sum(axis = 1, keepdims = True)

    def select_better(self, old, new):
        return max(old, new)

class F1(Metric):

    def step(self, output_data):
        loss, inputs, outputs, labels = output_data
        return outputs, labels

    def reduce(self, memory, config):
        outputs = [entry[0] for entry in memory]
        labels = [entry[1] for entry in memory]
        outputs = torch.argmax(torch.cat(outputs), dim = 1).cpu().numpy()
        labels = torch.cat(labels).cpu().numpy()
        score = sk_metrics.f1_score(labels, outputs)
        return score

    def select_better(self, old, new):
        return max(old, new)



from tqdm import tqdm
import rubik.utils as utils
import rubik.logger as logger

class Cube():
    '''
    Rubik's Cube
    '''

    def __init__(self):
        self.__init_config()
        self.__set_core_function()
        self.__set_default_action()
        self.__init_checklist()
        self.__init_other()

    def __init_config(self):
        '''
        Config includes ``func``: customizable functions, ``hparam``: \
        training hyperparameters, and ``data``: training related variables
        '''
        config = {}
        config['proc'] = {} # includes all variable functions
        config['hparam'] = {} # hyperparameters
        config['data'] = {} # datalodaers, current epoch, etc.
        self.config = config


    # Checklist before start training
    def __init_checklist(self):
        '''
        create a checklist 
        '''
        steps = ['hparam', 'dataloader', 'model', 'train', 'val', 
                 'optimizer', 'criterion', 'checkpoint']
        values = [0] * len(steps)
        self.checklist = dict(zip(steps, values))

    def __check(self, item_name):
        if item_name in self.checklist.keys():
            self.checklist[item_name] = 1
        
    def __checklist_complete(self):
        if sum(self.checklist.values()) != len(self.checklist):
            missing_list = []
            for k, v in self.checklist.items():
                if v == 0:
                    missing_list.append(k)
            raise Exception(f'{missing_list} not defined.')

    # Function setup
    def __set_core_function(self):
        self.config['proc']['loop'] = self.__loop
        self.config['proc']['train_wrapper'] = self.__train_wrapper
        self.config['proc']['train_loop'] = self.__train_loop
        self.config['proc']['val_wrapper'] = self.__val_wrapper
        self.config['proc']['val_loop'] = self.__val_loop
        self.config['proc']['checkpoint_wrapper'] = \
                                                self.__checkpoint_wrapper

    # Other initalization
    def __init_other(self):
        # init device
        self.config['data']['device'], _ = utils.get_device()

    # Before run starts
    def __init_run(self):
        # init run dir
        base_dir = './runs'
        run_name = f"{self.config['hparam']['run_name']}_{utils.get_time()}"
        run_dir = f'{base_dir}/{run_name}'
        checkpoint_dir = f'{run_dir}/checkpoints'
        utils.create_dir(base_dir)
        utils.create_dir(run_dir)
        utils.create_dir(checkpoint_dir)
        self.config['data']['run_dir'] = run_dir
        self.config['data']['checkpoint_dir'] = checkpoint_dir
        # init logger
        self.config['data']['logger'] = logger.Logger(self.config)
        # log training device
        _, device_log = utils.get_device()
        self.config['data']['logger'].log_message(device_log, 0)

    # Action decorator
    def set_action(self, action_path):
        def wrapper_set_action(func):
            self.config['proc'][action_path] = func
        return wrapper_set_action

    # Action does nothing by default
    def __set_default_action(self):
        action_sections = ['run', 'loop', 'epoch', 'batch', 'train', 'val', 
                           'train_batch', 'val_batch']
        no_action = lambda x : None
        for sec in action_sections:
            self.config['proc'][f'pre_{sec}'] = no_action
            self.config['proc'][f'post_{sec}'] = no_action
            

    # Training functions 
    def __loop(self, config):
        self.config['proc']['pre_loop'](self.config)
        for epoch in range(self.config['hparam']['epochs']):
            self.config['data']['current_epoch'] = epoch
            self.config['proc']['pre_epoch'](self.config)
            self.config['proc']['train_wrapper'](self.config)
            if self.__is_now(epoch, self.config['hparam']['val_interval']):
                self.config['proc']['val_wrapper'](self.config)
            if self.__is_now(epoch, self.config['hparam']['save_interval']):
                self.config['proc']['checkpoint_wrapper'](self.config)
            self.config['proc']['post_epoch'](self.config)
        self.config['proc']['post_loop'](self.config)

    def __train_wrapper(self, config):
        self.config['proc']['pre_train'](self.config)
        self.config['proc']['train_loop'](self.config)
        self.config['proc']['post_train'](self.config)
        for metric in self.config['data']['train_metrics']:
            result = metric.reduce_wrap(self.config)
            self.config['data']['logger'].log_value(
                f'train_{metric.__class__.__name__}', 
                result, 
                self.__total_iteration(),
                self.config['data']['current_epoch'])

    def __train_loop(self, config):
        for i, data in enumerate(tqdm(self.config['data']['train_loader'])):
            self.config['data']['current_batch'] = i
            data = self.__transfer_device(data)
            self.config['proc']['pre_train_batch'](self.config)
            outputs = self.config['proc']['train_func'](data, self.config)
            self.config['proc']['post_train_batch'](self.config)
            for metric in self.config['data']['train_metrics']:
                metric.step_wrap(outputs)

    def __val_wrapper(self, config):
        self.config['proc']['pre_val'](self.config)
        self.config['proc']['val_loop'](self.config)
        self.config['proc']['post_val'](self.config)
        for metric in self.config['data']['val_metrics']:
            result = metric.reduce_wrap(self.config)
            self.config['data']['logger'].log_value(
                f'val_{metric.__class__.__name__}', 
                result, 
                self.__total_iteration(),
                self.config['data']['current_epoch'])

    def __val_loop(self, config):
        for i, data in enumerate(tqdm(self.config['data']['val_loader'])):
            self.config['data']['cur_val_batch'] = i
            data = self.__transfer_device(data)
            self.config['proc']['pre_val_batch'](self.config)
            outputs = self.config['proc']['val_func'](data, self.config)
            self.config['proc']['post_val_batch'](self.config)
            for metric in self.config['data']['val_metrics']:
                metric.step_wrap(outputs)

    def __total_iteration(self):
        train_len = len(self.config['data']['train_loader'].dataset)
        base = train_len * self.config['data']['current_epoch']
        iterations = self.config['data']['current_batch'] * \
                     self.config['hparam']['batch_size']
        return base + iterations

    def __transfer_device(self, data):
        data_batch = []
        for data_i in data:
            data_batch.append(data_i.to(self.config['data']['device']))
        return data_batch

    def __checkpoint_wrapper(self, config):
        self.config['proc']['checkpoint'](self.config)

    def __is_now(self, current, interval):
        return (current + 1) % interval == 0


    # Required setup steps
    def set_hparam(self, hparam_func):
        self.config['hparam'] = hparam_func(self.config)
        self.__check('hparam')
    
    #TODO override to support multiple test loader
    def set_dataloader(self, loader_func):
        train_loader, test_loader = loader_func(self.config)
        self.config['data']['train_loader'] = train_loader
        self.config['data']['val_loader'] = test_loader
        self.__check('dataloader')

    def set_model(self, model_func):
        model = model_func(self.config)
        model = model.to(self.config['data']['device'])
        self.config['data']['model'] = model
        self.__check('model')

    def set_optimizer(self, optimizer_func):
        self.config['data']['optimizer'] = optimizer_func(self.config)
        self.__check('optimizer')

    def set_criterion(self, criterion_func):
        self.config['data']['criterion'] = criterion_func(self.config)
        self.__check('criterion')

    def set_metrics(self, metrics_func):
        train_metrics, val_metrics = metrics_func(self.config)
        self.config['data']['train_metrics'] = train_metrics
        self.config['data']['val_metrics'] = val_metrics
        self.__check('criterion')

    def set_train(self, train_func):
        self.config['proc']['train_func'] = train_func
        self.__check('train')

    def set_val(self, val_func):
        self.config['proc']['val_func'] = val_func
        self.__check('val')

    def set_checkpoint(self, checkpoint_func):
        self.config['proc']['checkpoint'] = checkpoint_func
        self.__check('checkpoint')

    # Start training
    def start_run(self):
        self.config['proc']['pre_run'](self.config)
        self.__checklist_complete()
        self.__init_run()
        self.config['proc']['loop'](self.config)
        self.config['proc']['post_run'](self.config)



import unittest
from unittest import TestCase

import rubik.rubik as rubik

import torch
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import Dataset

@unittest.skip('tested')
class TestRubic(TestCase):

    @classmethod
    def setUpClass(self):
        config = {}
        config['input_size'] = 5
        config['hidden_size'] = 32
        config['num_layers'] = 2
        config['dropout'] = 0.2
        self.rubik= rubik.Rubik()
    
    @unittest.skip('legacy test')
    def test_pipeline(self):
        #import pdb; pdb.set_trace()
        def train_func(data, config):
            #print('training')
            pass
        def val_func(data, config):
            #print('val')
            pass

        @self.rubik.set_action('pre_train')
        def pre_loop(config):
            print(config['hparam']['epochs'])

        self.rubik.set_hparam({'epochs':10})
        self.rubik.set_train(train_func)
        self.rubik.set_val(val_func)

        transform = transforms.Compose([transforms.ToTensor(),
                                    transforms.Normalize((0.5, 0.5, 0.5), 
                                                         (0.5, 0.5, 0.5))])
        CIFAR10 = torchvision.datasets.CIFAR10
        trainset = CIFAR10(root='./data', train=False,
                       download=True, transform=transform)
        trainloader = torch.utils.data.DataLoader(trainset, 
                                            batch_size=128, 
                                            shuffle=True, 
                                            num_workers=8,
                                            pin_memory=True) 
        self.rubik.set_dataloader(trainloader, trainloader)

        model = torchvision.models.resnet18()
        self.rubik.set_model(model)

        self.rubik.start()




import os
import csv
import torch

from PIL import Image

from torchvision.utils import make_grid, save_image

from torch.utils.tensorboard import SummaryWriter
try:
    from torch.utils.tensorboard.summary import hparams
except:
    hparams = lambda x, y: (0, 0, 0)


# Epoch for message logging and iteration for value logging
class Logger():

    def __init__(self, config, cli_log = True, f_log = True, tb_log = True):
        ''' CLI, file and tensorboard logger '''
        self.cli = CLILogger(cli_log)
        self.f = FileLogger(config['data']['run_dir'], f_log)
        self.tb = TensorBoardLogger(config['data']['run_dir'], tb_log)
        self.loggers = [self.f, self.tb, self.cli]

    def log_images(self, img_name, images):
        if images.shape[1] > 3:
            images = images.sum(dim = 1, keepdim = True)
        for logger in self.loggers:
            logger.log_images(img_name, images)

    def log_message(self, message, epoch):
        for logger in self.loggers:
            logger.log_message(message, epoch)

    def log_value(self, value_name, value, iteration, epoch):
        for logger in self.loggers:
            logger.log_value(value_name, value, iteration, epoch)

    def log_hparams(self, hparam, metrics):
        for logger in self.loggers:
            logger.log_hparams(hparam, metrics)

    def log_graph(self, model, data):
        self.tb.log_graph(model, data)

    # Extra layer has to have string name of the layer
    def log_grad(self, model, iteration, num_layer = 3, extra_layers = []):
        selected_layers = self.select_layers(model, num_layer, extra_layers)
        for name, parameters in model.named_parameters():
            true_name = name.split('.')[0]
            if true_name in selected_layers:
                self.tb.log_grad(name, parameters.grad, iteration)

    def select_layers(self, model, num_layer, extra_layers):
        all_layers = []
        for name, _ in model.named_parameters():
            true_name = name.split('.')[0]
            if true_name not in all_layers: all_layers.append(true_name)

        # Select layers according to layers
        selected_layers = []
        for i in range(num_layer - 1):
            factor = int(len(all_layers) / (num_layer - 1))
            selected_layers.append(all_layers[i * factor])
        selected_layers.append(all_layers[-1])
        selected_layers.extend(extra_layers)
        return selected_layers

    def close(self):
        self.tb.close()

class CLILogger():
    def __init__(self, enable = True):
        self.disable = not enable

    def log_images(self, img_name, images):
        if self.disable: return
        print('Image {}.png is saved.'.format(img_name))

    def log_message(self, message, epoch):
        if self.disable: return
        print('epoch: {}, msg: {}'.format(epoch, message))

    def log_value(self, value_name, value, iteration, epoch):
        if self.disable: return
        print('ep: {}, iter: {}, {}: {}'.format(epoch, iteration, 
                                                value_name, value))

    def log_hparams(self, hparam, metrics):
        if self.disable: return
        if not metrics is None:
            hparam.update(metrics)
        print('Hyperparameter: \n {}'.format(hparam))


class FileLogger():
    def __init__(self, log_dir, enable = True):
        self.disable = not enable
        self.metrics_dicts = []
        self.metrics_column = ['epoch', 'iteration'] # Avoid repetitive add
        self.metrics_file = self.init_file('metrics.csv', log_dir)
        self.message_file = self.init_file('log.txt', log_dir)
        self.hparam_file = self.init_file('hparam.csv', log_dir)
        self.data_dir = self.init_dir('data', log_dir)

    def init_file(self, file_name, save_dir):
        if self.disable: return
        file_dir = '{}/{}'.format(save_dir, file_name)
        if os.path.exists(file_dir):
            raise Exception('File {} exists!'.format(file_dir))
        return file_dir
        
    def init_dir(self, dir_name, save_dir):
        if self.disable: return
        sub_dir = '{}/{}'.format(save_dir, dir_name)
        if os.path.exists(sub_dir):
            raise Exception('Path {} exists!'.format(sub_dir))
        else:
            os.mkdir(sub_dir)
        return sub_dir
        
    def log_images(self, img_name, images):
        if self.disable: return
        img_grid = make_grid(images)
        save_image(img_grid, '{}/{}.png'.format(self.data_dir, img_name))

    def log_message(self, message, epoch):
        if self.disable: return
        with open(self.message_file, 'a') as f:
            f.write('epoch: {} | {}\n'.format(epoch, message))

    def log_value(self, value_name, value, iteration, epoch):
        if self.disable: return
        m_dicts = self.metrics_dicts
        if len(m_dicts) == 0 or m_dicts[-1]['iteration'] != iteration:
            m_dicts.append({'epoch' : epoch,
                            'iteration' : iteration,
                            value_name : value})
        else:
            m_dicts[-1][value_name] = value

        if value_name not in self.metrics_column: 
            self.metrics_column.append(value_name)
        with open(self.metrics_file, 'w') as f:
            writer = csv.DictWriter(f, 
                                    fieldnames = self.metrics_column, 
                                    restval = 'NaN')
            writer.writeheader()
            for data in m_dicts:
                writer.writerow(data)

    def log_hparams(self, hparam, metrics):
        if self.disable: return
        if not metrics is None:
            hparam.update(metrics)
        with open(self.hparam_file, 'w') as f:
            writer = csv.writer(f)
            for key, value in hparam.items():
                writer.writerow([key, value])


class TensorBoardLogger():

    def __init__(self, log_dir, enable = True):
        self.writer = ModSummaryWriter(log_dir)
        self.disable = not enable

    def log_images(self, img_name, images):
        if self.disable: return
        img_grid = make_grid(images)
        self.writer.add_image(img_name, img_grid)

    def log_value(self, value_name, value, iteration, epoch):
        if self.disable: return
        self.writer.add_scalar(value_name, value, iteration)

    def log_message(self, message, epoch):
        if self.disable: return
        self.writer.add_text('log', message, epoch)

    def log_hparams(self, hparam, metrics):
        if self.disable: return
        self.writer.add_mod_hparams(hparam, metrics)

    def log_graph(self, model, data):
        if self.disable: return
        self.writer.add_graph(model, data)

    def log_grad(self, name, data, iteration):
        if self.disable: return
        self.writer.add_histogram(name, data, iteration)

    def close(self):
        self.writer.close()

# Custom Hparam Writer that prevent split folders
# https://github.com/pytorch/pytorch/issues/32651#issuecomment-643791116
class ModSummaryWriter(SummaryWriter):
    def add_mod_hparams(self, hparam_dict, metric_dict):
        torch._C._log_api_usage_once("tensorboard.logging.add_hparams")
        if type(hparam_dict) is not dict or type(metric_dict) is not dict:
            raise TypeError('hparam_dict and metric_dict should be dictionary.')
        exp, ssi, sei = hparams(hparam_dict, metric_dict)
        logdir = self._get_file_writer().get_logdir()
        self.file_writer.add_summary(exp)
        self.file_writer.add_summary(ssi)
        self.file_writer.add_summary(sei)
        for k, v in metric_dict.items():
            self.add_scalar(k, v)



import unittest
from unittest import TestCase

import metrics
import random

import torch
import torchvision
import torchvision.transforms as transforms

#@unittest.skip('tested')
class TestMetrics(TestCase):

    @classmethod
    def setUpClass(self):
        pass
    
    #@unittest.skip('tested')
    def test_accuracy(self):
        acc = metrics.Accuracy()
        for i in range(10):
            for j in range(200):
                bs = 10
                loss = 0
                inputs = torch.rand(bs, 3, 32, 32)
                outputs = torch.rand(bs, 1)
                labels = torch.randint(2 + i, [bs, 1])
                outputs = (loss, inputs, outputs, labels)
                acc.step_wrap(outputs)
            acc_score = acc.reduce_wrap(None)
            self.assertTrue(acc_score < 1 / (i + 2) + 0.1)
            self.assertTrue(acc_score > 1 / (i + 2) - 0.1)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rubik-pytorch", 
    version="0.0.2",
    author="Jimin Tan",
    author_email="tanjimin@nyu.edu",
    description="Highly customizable Deep Learning training framework \
        base on PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tanjimin/rubik",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import torch
import datetime, os

def get_device():
    message = ''
    # Check device
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    message += "Computation device: {}".format(device) + '\n '
    #Additional Info when using cuda
    if device.type == 'cuda':
        message += str(torch.cuda.get_device_name(0)) + '\n'
        message += ' Memory:\n Allocated:{} GB\n Cached:{} GB\n'.format(
            round(torch.cuda.memory_allocated(0)/1024**3,1),
            round(torch.cuda.memory_cached(0)/1024**3,1))
    return device, message

def get_time():
    return datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')[:-3]

def create_dir(path):
    if not os.path.exists(path):
        os.mkdir(path)
